@@ -1,6 +1,6 @@
 /* -*- mode: C; mode: fold; -*- */
 /*
-Copyright (C) 2005 John E. Davis
+Copyright (C) 2005, 2006, 2007 John E. Davis
 
 This file is part of the S-Lang Curl Module
 
@@ -31,6 +31,39 @@ USA.
 SLANG_MODULE(curl);
 #include "version.h"
 
+/* Unfortunately symbols such as CURLOPT_FTP_ACCOUNT are not macros and the 
+ * only way to test for their existence is through the version number. Sigh.
+ */
+#define MAKE_CURL_VERSION(a,b,c) (((a)<<16) + ((b)<<8) + (c))
+#define CURL_VERSION_GE(a,b,c) \
+   (LIBCURL_VERSION_NUM >= MAKE_CURL_VERSION((a),(b),(c)))
+#define CURL_VERSION_LT(a,b,c) \
+   (LIBCURL_VERSION_NUM < MAKE_CURL_VERSION((a),(b),(c)))
+
+/* Unfortunately symbols appear and disappear between various curl releases,
+ * and because those symbols are enums, hacks such as the following are
+ * necessary.  This is why enums should not be used in this manner.
+ */
+#if CURL_VERSION_LT(7,16,0)
+# define HAVE_CURLOPT_PREQUOTE
+# define HAVE_CURLOPT_SOURCE_PREQUOTE
+# define HAVE_CURLOPT_POSTQUOTE
+# define HAVE_CURLOPT_SOURCE_POSTQUOTE
+# define HAVE_CURLOPT_SOURCE_USERPWD
+# define HAVE_CURLOPT_PREQUOTE
+# define HAVE_CURLOPT_USERPWD
+#endif
+
+#if CURL_VERSION_GE(7,14,0)
+# define HAVE_CURLOPT_FTP_ACCOUNT
+#endif
+
+#if CURL_VERSION_GE(7,14,0) && CURL_VERSION_LT(7,16,0)
+# define HAVE_CURLOPT_SOURCE_QUOTE
+# define HAVE_CURLOPT_SOURCE_URL
+#endif
+  
+
 static int Curl_Error = 0;
 static SLtype Easy_Type_Id = 0;
 static SLtype Multi_Type_Id = 0;
@@ -291,7 +324,7 @@ static int check_handle (Easy_Type *ez, unsigned int flags)
      }
    if (ez->flags & flags)
      {
-	SLang_verror (SL_RunTime_Error, "It is illegal to call this function when curl_perform is running");
+	SLang_verror (SL_RunTime_Error, "It is illegal to call this function while curl_perform is running");
 	return -1;
      }
 
@@ -581,7 +614,9 @@ static int do_setopt (Easy_Type *ez, CURLoption opt, int nargs)
 	return set_long_opt (ez, opt, nargs, 0, 0L);   /* FIXME */
 
       case CURLOPT_NETRC_FILE:
+#ifdef HAVE_CURLOPT_USERPWD
       case CURLOPT_USERPWD:
+#endif
       case CURLOPT_PROXYUSERPWD:
 	return set_string_opt (ez, opt, nargs);
 
@@ -645,14 +680,16 @@ static int do_setopt (Easy_Type *ez, CURLoption opt, int nargs)
 	/* FTP options */
       case CURLOPT_FTPPORT:
 	return set_string_opt (ez, opt, nargs);
-
       case CURLOPT_QUOTE:	       /* FIXME: linked list */
 	return set_strlist_opt (ez, opt, nargs, &ez->quote);
+#ifdef HAVE_CURLOPT_POSTQUOTE
       case CURLOPT_POSTQUOTE:
 	return set_strlist_opt (ez, opt, nargs, &ez->postquote);
+#endif
+#ifdef HAVE_CURLOPT_PREQUOTE
       case CURLOPT_PREQUOTE:
 	return set_strlist_opt (ez, opt, nargs, &ez->prequote);
-
+#endif
       case CURLOPT_FTPLISTONLY:
       case CURLOPT_FTPAPPEND:
       case CURLOPT_FTP_USE_EPRT:
@@ -664,30 +701,30 @@ static int do_setopt (Easy_Type *ez, CURLoption opt, int nargs)
       case CURLOPT_FTP_SSL:
 	return set_long_opt (ez, opt, nargs, 0, 0L);   /* FIXME: specific values */
 
-#ifdef CURLOPT_SOURCE_URL
+#ifdef HAVE_CURLOPT_SOURCE_URL
       case CURLOPT_SOURCE_URL:
 	return set_string_opt (ez, opt, nargs);
 #endif
-#ifdef CURLOPT_SOURCE_USERPWD
+#ifdef HAVE_CURLOPT_SOURCE_USERPWD
       case CURLOPT_SOURCE_USERPWD:
 	return set_string_opt (ez, opt, nargs);
 #endif
-#ifdef CURLOPT_SOURCE_QUOTE
+#ifdef HAVE_CURLOPT_SOURCE_QUOTE
       case CURLOPT_SOURCE_QUOTE:
 	return set_strlist_opt (ez, opt, nargs, &ez->source_quote);
 #endif
-#ifdef CURLOPT_SOURCE_PREQUOTE
+#ifdef HAVE_CURLOPT_SOURCE_PREQUOTE
       case CURLOPT_SOURCE_PREQUOTE:
 	return set_strlist_opt (ez, opt, nargs, &ez->source_prequote);
 #endif
-#ifdef CURLOPT_SOURCE_POSTQUOTE
+#ifdef HAVE_CURLOPT_SOURCE_POSTQUOTE
       case CURLOPT_SOURCE_POSTQUOTE:
 	return set_strlist_opt (ez, opt, nargs, &ez->source_postquote);
 #endif
-
+#ifdef HAVE_CURLOPT_FTP_ACCOUNT
       case CURLOPT_FTP_ACCOUNT:
 	return set_string_opt (ez, opt, nargs);
-
+#endif
 	/* protocol options */
       case CURLOPT_TRANSFERTEXT:
       case CURLOPT_CRLF:
@@ -1443,6 +1480,96 @@ static int get_multi_length_intrin (void)
 
 /*}}}*/
 
+
+static void escape_intrin (SLang_BString_Type *bstr)
+{
+   SLang_MMT_Type *mmt;
+   Easy_Type *ez;
+   char *escaped_string;
+   char *url;
+   unsigned int len;
+
+   if (NULL == (url = (char *) SLbstring_get_pointer (bstr, &len)))
+     return;
+
+   if (NULL == (mmt = pop_easy_type (&ez, 0))) 
+     return;
+
+#if CURL_VERSION_GE(7,15,4)
+   escaped_string = curl_easy_escape (ez->handle, url, (int)len);
+#else
+   escaped_string = curl_escape (url, (int) len);
+#endif
+
+   if (escaped_string == NULL)
+     SLang_set_error (Curl_Error);
+   else
+     {
+	char *str = SLang_create_slstring (escaped_string);
+	if (str != NULL)
+	  {
+	     (void) SLang_push_string (str);
+	     SLang_free_slstring (str);
+	  }
+	curl_free (escaped_string);
+     }
+   SLang_free_mmt(mmt);
+}
+
+static void unescape_intrin (char *url)
+{
+   SLang_MMT_Type *mmt;
+   Easy_Type *ez;
+   char *unescaped_string;
+#if CURL_VERSION_GE(7,15,4)
+   int outlength;
+#endif
+
+   if (NULL == (mmt = pop_easy_type (&ez, 0))) 
+     return;
+
+#if CURL_VERSION_GE(7,15,4)
+   unescaped_string = curl_easy_unescape (ez->handle, url, 0, &outlength);
+#else
+   unescaped_string = curl_unescape (url, 0);
+#endif
+
+   if (unescaped_string == NULL)
+     SLang_set_error (Curl_Error);
+   else
+     {
+	int len = strlen (unescaped_string);
+#if CURL_VERSION_GE(7,15,4)
+	if (len != outlength)
+	  {
+	     SLang_BString_Type *bstr = SLbstring_create ((unsigned char *)unescaped_string, outlength);
+	     if (bstr != NULL)
+	       {
+		  (void) SLang_push_bstring (bstr);
+		  SLbstring_free (bstr);
+	       }
+	  }
+	else
+#endif
+	  {
+	     char *str = SLang_create_nslstring (unescaped_string, len);
+	     if (str != NULL)
+	       {
+		  (void) SLang_push_string (str);
+		  SLang_free_slstring (str);
+	       }
+	  }
+	curl_free (unescaped_string);
+     }
+   SLang_free_mmt(mmt);
+}
+
+static char *easy_strerror_intrin (int *codep)
+{
+   return (char *) curl_easy_strerror ((CURLcode) *codep);
+}
+
+
 static SLang_Intrin_Fun_Type Module_Intrinsics [] =
 {
    MAKE_INTRINSIC_1("curl_new", new_curl_intrin, SLANG_VOID_TYPE, SLANG_STRING_TYPE),
@@ -1460,6 +1587,12 @@ static SLang_Intrin_Fun_Type Module_Intrinsics [] =
    MAKE_INTRINSIC_0("curl_multi_close", multi_close, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("curl_multi_info_read", multi_info_read, SLANG_VOID_TYPE),
    
+   MAKE_INTRINSIC_I("curl_easy_strerror", easy_strerror_intrin, SLANG_STRING_TYPE),
+   MAKE_INTRINSIC_I("curl_strerror", easy_strerror_intrin, SLANG_STRING_TYPE),
+
+   MAKE_INTRINSIC_1("curl_easy_escape", escape_intrin, SLANG_VOID_TYPE, SLANG_BSTRING_TYPE),
+   MAKE_INTRINSIC_1("curl_easy_unescape", unescape_intrin, SLANG_VOID_TYPE, SLANG_STRING_TYPE),
+
    /* Local Additions */
    MAKE_INTRINSIC_0("curl_get_url", get_url_intrin, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("curl_multi_length", get_multi_length_intrin, SLANG_INT_TYPE),
@@ -1537,8 +1670,12 @@ static SLang_IConstant_Type Module_IConstants [] =
    MAKE_ICONSTANT("CURLOPT_HTTP_VERSION", CURLOPT_HTTP_VERSION),
    MAKE_ICONSTANT("CURLOPT_FTPPORT", CURLOPT_FTPPORT),
    MAKE_ICONSTANT("CURLOPT_QUOTE", CURLOPT_QUOTE),
+#ifdef HAVE_CURLOPT_POSTQUOTE
    MAKE_ICONSTANT("CURLOPT_POSTQUOTE", CURLOPT_POSTQUOTE),
+#endif
+#ifdef HAVE_CURLOPT_PREQUOTE
    MAKE_ICONSTANT("CURLOPT_PREQUOTE", CURLOPT_PREQUOTE),
+#endif
    MAKE_ICONSTANT("CURLOPT_FTPLISTONLY", CURLOPT_FTPLISTONLY),
    MAKE_ICONSTANT("CURLOPT_FTPAPPEND", CURLOPT_FTPAPPEND),
    MAKE_ICONSTANT("CURLOPT_FTP_USE_EPRT", CURLOPT_FTP_USE_EPRT),
@@ -1546,22 +1683,24 @@ static SLang_IConstant_Type Module_IConstants [] =
    MAKE_ICONSTANT("CURLOPT_FTP_CREATE_MISSING_DIRS", CURLOPT_FTP_CREATE_MISSING_DIRS),
    MAKE_ICONSTANT("CURLOPT_FTP_RESPONSE_TIMEOUT", CURLOPT_FTP_RESPONSE_TIMEOUT),
    MAKE_ICONSTANT("CURLOPT_FTP_SSL", CURLOPT_FTP_SSL),
-#ifdef CURLOPT_SOURCE_URL
+#ifdef HAVE_CURLOPT_SOURCE_URL
    MAKE_ICONSTANT("CURLOPT_SOURCE_URL", CURLOPT_SOURCE_URL),
 #endif
-#ifdef CURLOPT_SOURCE_USERPWD
+#ifdef HAVE_CURLOPT_SOURCE_USERPWD
    MAKE_ICONSTANT("CURLOPT_SOURCE_USERPWD", CURLOPT_SOURCE_USERPWD),
 #endif
-#ifdef CURLOPT_SOURCE_QUOTE
+#ifdef HAVE_CURLOPT_SOURCE_QUOTE
    MAKE_ICONSTANT("CURLOPT_SOURCE_QUOTE", CURLOPT_SOURCE_QUOTE),
 #endif
-#ifdef CURLOPT_SOURCE_PREQUOTE
+#ifdef HAVE_CURLOPT_SOURCE_PREQUOTE
    MAKE_ICONSTANT("CURLOPT_SOURCE_PREQUOTE", CURLOPT_SOURCE_PREQUOTE),
 #endif
-#ifdef CURLOPT_SOURCE_POSTQUOTE
+#ifdef HAVE_CURLOPT_SOURCE_POSTQUOTE
    MAKE_ICONSTANT("CURLOPT_SOURCE_POSTQUOTE", CURLOPT_SOURCE_POSTQUOTE),
 #endif
+#ifdef HAVE_CURLOPT_FTP_ACCOUNT
    MAKE_ICONSTANT("CURLOPT_FTP_ACCOUNT", CURLOPT_FTP_ACCOUNT),
+#endif
    MAKE_ICONSTANT("CURLOPT_TRANSFERTEXT", CURLOPT_TRANSFERTEXT),
    MAKE_ICONSTANT("CURLOPT_CRLF", CURLOPT_CRLF),
    MAKE_ICONSTANT("CURLOPT_RANGE", CURLOPT_RANGE),
@@ -1611,6 +1750,10 @@ static SLang_IConstant_Type Module_IConstants [] =
    MAKE_ICONSTANT("CURL_GLOBAL_WIN32", CURL_GLOBAL_WIN32),
    MAKE_ICONSTANT("CURL_GLOBAL_NOTHING", CURL_GLOBAL_NOTHING),
 
+   MAKE_ICONSTANT("CURL_NETRC_IGNORED", CURL_NETRC_IGNORED),
+   MAKE_ICONSTANT("CURL_NETRC_OPTIONAL", CURL_NETRC_OPTIONAL),
+   MAKE_ICONSTANT("CURL_NETRC_REQUIRED", CURL_NETRC_REQUIRED),
+
    MAKE_ICONSTANT("CURLINFO_EFFECTIVE_URL", CURLINFO_EFFECTIVE_URL),
    MAKE_ICONSTANT("CURLINFO_RESPONSE_CODE", CURLINFO_RESPONSE_CODE),
    MAKE_ICONSTANT("CURLINFO_TOTAL_TIME", CURLINFO_TOTAL_TIME),
@@ -1639,6 +1782,76 @@ static SLang_IConstant_Type Module_IConstants [] =
    MAKE_ICONSTANT("CURLINFO_NUM_CONNECTS", CURLINFO_NUM_CONNECTS),
    MAKE_ICONSTANT("CURLINFO_SSL_ENGINES", CURLINFO_SSL_ENGINES),
 
+   MAKE_ICONSTANT("CURLE_OK", CURLE_OK),
+   MAKE_ICONSTANT("CURLE_UNSUPPORTED_PROTOCOL", CURLE_UNSUPPORTED_PROTOCOL),
+   MAKE_ICONSTANT("CURLE_FAILED_INIT", CURLE_FAILED_INIT),
+   MAKE_ICONSTANT("CURLE_URL_MALFORMAT", CURLE_URL_MALFORMAT),
+   MAKE_ICONSTANT("CURLE_URL_MALFORMAT_USER", CURLE_URL_MALFORMAT_USER),
+   MAKE_ICONSTANT("CURLE_COULDNT_RESOLVE_PROXY", CURLE_COULDNT_RESOLVE_PROXY),
+   MAKE_ICONSTANT("CURLE_COULDNT_RESOLVE_HOST", CURLE_COULDNT_RESOLVE_HOST),
+   MAKE_ICONSTANT("CURLE_COULDNT_CONNECT", CURLE_COULDNT_CONNECT),
+   MAKE_ICONSTANT("CURLE_FTP_WEIRD_SERVER_REPLY", CURLE_FTP_WEIRD_SERVER_REPLY),
+   MAKE_ICONSTANT("CURLE_FTP_ACCESS_DENIED", CURLE_FTP_ACCESS_DENIED),
+   MAKE_ICONSTANT("CURLE_FTP_USER_PASSWORD_INCORRECT", CURLE_FTP_USER_PASSWORD_INCORRECT),
+   MAKE_ICONSTANT("CURLE_FTP_WEIRD_PASS_REPLY", CURLE_FTP_WEIRD_PASS_REPLY),
+   MAKE_ICONSTANT("CURLE_FTP_WEIRD_USER_REPLY", CURLE_FTP_WEIRD_USER_REPLY),
+   MAKE_ICONSTANT("CURLE_FTP_WEIRD_PASV_REPLY", CURLE_FTP_WEIRD_PASV_REPLY),
+   MAKE_ICONSTANT("CURLE_FTP_WEIRD_227_FORMAT", CURLE_FTP_WEIRD_227_FORMAT),
+   MAKE_ICONSTANT("CURLE_FTP_CANT_GET_HOST", CURLE_FTP_CANT_GET_HOST),
+   MAKE_ICONSTANT("CURLE_FTP_CANT_RECONNECT", CURLE_FTP_CANT_RECONNECT),
+   MAKE_ICONSTANT("CURLE_FTP_COULDNT_SET_BINARY", CURLE_FTP_COULDNT_SET_BINARY),
+   MAKE_ICONSTANT("CURLE_PARTIAL_FILE", CURLE_PARTIAL_FILE),
+   MAKE_ICONSTANT("CURLE_FTP_COULDNT_RETR_FILE", CURLE_FTP_COULDNT_RETR_FILE),
+   MAKE_ICONSTANT("CURLE_FTP_WRITE_ERROR", CURLE_FTP_WRITE_ERROR),
+   MAKE_ICONSTANT("CURLE_FTP_QUOTE_ERROR", CURLE_FTP_QUOTE_ERROR),
+   MAKE_ICONSTANT("CURLE_HTTP_RETURNED_ERROR", CURLE_HTTP_RETURNED_ERROR),
+   MAKE_ICONSTANT("CURLE_WRITE_ERROR", CURLE_WRITE_ERROR),
+   MAKE_ICONSTANT("CURLE_MALFORMAT_USER", CURLE_MALFORMAT_USER),
+   MAKE_ICONSTANT("CURLE_FTP_COULDNT_STOR_FILE", CURLE_FTP_COULDNT_STOR_FILE),
+   MAKE_ICONSTANT("CURLE_READ_ERROR", CURLE_READ_ERROR),
+   MAKE_ICONSTANT("CURLE_OUT_OF_MEMORY", CURLE_OUT_OF_MEMORY),
+   MAKE_ICONSTANT("CURLE_OPERATION_TIMEOUTED", CURLE_OPERATION_TIMEOUTED),
+   MAKE_ICONSTANT("CURLE_FTP_COULDNT_SET_ASCII", CURLE_FTP_COULDNT_SET_ASCII),
+   MAKE_ICONSTANT("CURLE_FTP_PORT_FAILED", CURLE_FTP_PORT_FAILED),
+   MAKE_ICONSTANT("CURLE_FTP_COULDNT_USE_REST", CURLE_FTP_COULDNT_USE_REST),
+   MAKE_ICONSTANT("CURLE_FTP_COULDNT_GET_SIZE", CURLE_FTP_COULDNT_GET_SIZE),
+   MAKE_ICONSTANT("CURLE_HTTP_RANGE_ERROR", CURLE_HTTP_RANGE_ERROR),
+   MAKE_ICONSTANT("CURLE_HTTP_POST_ERROR", CURLE_HTTP_POST_ERROR),
+   MAKE_ICONSTANT("CURLE_SSL_CONNECT_ERROR", CURLE_SSL_CONNECT_ERROR),
+   MAKE_ICONSTANT("CURLE_BAD_DOWNLOAD_RESUME", CURLE_BAD_DOWNLOAD_RESUME),
+   MAKE_ICONSTANT("CURLE_FILE_COULDNT_READ_FILE", CURLE_FILE_COULDNT_READ_FILE),
+   MAKE_ICONSTANT("CURLE_LDAP_CANNOT_BIND", CURLE_LDAP_CANNOT_BIND),
+   MAKE_ICONSTANT("CURLE_LDAP_SEARCH_FAILED", CURLE_LDAP_SEARCH_FAILED),
+   MAKE_ICONSTANT("CURLE_LIBRARY_NOT_FOUND", CURLE_LIBRARY_NOT_FOUND),
+   MAKE_ICONSTANT("CURLE_FUNCTION_NOT_FOUND", CURLE_FUNCTION_NOT_FOUND),
+   MAKE_ICONSTANT("CURLE_ABORTED_BY_CALLBACK", CURLE_ABORTED_BY_CALLBACK),
+   MAKE_ICONSTANT("CURLE_BAD_FUNCTION_ARGUMENT", CURLE_BAD_FUNCTION_ARGUMENT),
+   MAKE_ICONSTANT("CURLE_BAD_CALLING_ORDER", CURLE_BAD_CALLING_ORDER),
+   MAKE_ICONSTANT("CURLE_INTERFACE_FAILED", CURLE_INTERFACE_FAILED),
+   MAKE_ICONSTANT("CURLE_BAD_PASSWORD_ENTERED", CURLE_BAD_PASSWORD_ENTERED),
+   MAKE_ICONSTANT("CURLE_TOO_MANY_REDIRECTS", CURLE_TOO_MANY_REDIRECTS ),
+   MAKE_ICONSTANT("CURLE_UNKNOWN_TELNET_OPTION", CURLE_UNKNOWN_TELNET_OPTION),
+   MAKE_ICONSTANT("CURLE_TELNET_OPTION_SYNTAX", CURLE_TELNET_OPTION_SYNTAX ),
+   MAKE_ICONSTANT("CURLE_OBSOLETE", CURLE_OBSOLETE),
+   MAKE_ICONSTANT("CURLE_SSL_PEER_CERTIFICATE", CURLE_SSL_PEER_CERTIFICATE),
+   MAKE_ICONSTANT("CURLE_GOT_NOTHING", CURLE_GOT_NOTHING),
+   MAKE_ICONSTANT("CURLE_SSL_ENGINE_NOTFOUND", CURLE_SSL_ENGINE_NOTFOUND),
+   MAKE_ICONSTANT("CURLE_SSL_ENGINE_SETFAILED", CURLE_SSL_ENGINE_SETFAILED),
+   MAKE_ICONSTANT("CURLE_SEND_ERROR", CURLE_SEND_ERROR),
+   MAKE_ICONSTANT("CURLE_RECV_ERROR", CURLE_RECV_ERROR),
+   MAKE_ICONSTANT("CURLE_SHARE_IN_USE", CURLE_SHARE_IN_USE),
+   MAKE_ICONSTANT("CURLE_SSL_CERTPROBLEM", CURLE_SSL_CERTPROBLEM),
+   MAKE_ICONSTANT("CURLE_SSL_CIPHER", CURLE_SSL_CIPHER),
+   MAKE_ICONSTANT("CURLE_SSL_CACERT", CURLE_SSL_CACERT),
+   MAKE_ICONSTANT("CURLE_BAD_CONTENT_ENCODING", CURLE_BAD_CONTENT_ENCODING),
+   MAKE_ICONSTANT("CURLE_LDAP_INVALID_URL", CURLE_LDAP_INVALID_URL),
+   MAKE_ICONSTANT("CURLE_FILESIZE_EXCEEDED", CURLE_FILESIZE_EXCEEDED),
+   MAKE_ICONSTANT("CURLE_FTP_SSL_FAILED", CURLE_FTP_SSL_FAILED),
+   MAKE_ICONSTANT("CURLE_SEND_FAIL_REWIND", CURLE_SEND_FAIL_REWIND),
+   MAKE_ICONSTANT("CURLE_SSL_ENGINE_INITFAILED", CURLE_SSL_ENGINE_INITFAILED),
+#if CURL_VERSION_GE(7,14,0)
+   MAKE_ICONSTANT("CURLE_LOGIN_DENIED", CURLE_LOGIN_DENIED),
+#endif
    SLANG_END_ICONST_TABLE
 };
 
